--
--  AY3819X.vhd
--
--  Top entity of AY3819X.
--
--        Copyright (C)2001 SEILEBOST
--                   All rights reserved.
--
-- $Id: AY3819.vhd, v0.3 2001/11/14 00:00:00 SEILEBOST $
--
-- TODO :
--  Many verification !!
-- Remark :
--  Two Clock period and an one clock period where CS = '0' between two commands !

library IEEE;
library UNISIM;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_STD.all;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use UNISIM.Vcomponents.ALL;

entity AY3819X is
    Port ( D       : inout  std_logic_vector(7 downto 0);
           RESET   : in     std_logic;
           CLOCK   : in     std_logic;
           BUSCTR  : in     std_logic_vector(2 downto 0);
           A8      : in     std_logic;
           A9      : in     std_logic;
           CS      : in     std_logic;
           IOA     : inout  std_logic_vector(7 downto 0);
           IOB     : inout  std_logic_vector(7 downto 0);
           AnalogA : out    std_logic;
           AnalogB : out    std_logic;
           AnalogC : out    std_logic );
end AY3819X;


architecture Behavioral of AY3819X is

 SIGNAL BUS_CS : std_logic_vector(15 downto 0);

 -- Create register 
 SIGNAL R0    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Fine Tune channel A
 SIGNAL R1    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Coarse Tune channel A
 SIGNAL R2    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Fine Tune  channel B
 SIGNAL R3    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Coarse Tune channel B
 SIGNAL R4    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Fine Tune channel C
 SIGNAL R5    : std_logic_vector(7  downto 0) := "00000000";  -- Tone generator frequency Coarse Tune channel B
 SIGNAL R6    : std_logic_vector(7  downto 0) := "00000000";  -- Noise generator frequency
 SIGNAL R7    : std_logic_vector(7  downto 0) := "00000000";  -- Mixer Control I/O Enable
 SIGNAL R8    : std_logic_vector(7  downto 0) := "00000000";  -- Amplitude control channel A
 SIGNAL R9    : std_logic_vector(7  downto 0) := "00000000";  -- Amplitude control channel B
 SIGNAL R10   : std_logic_vector(7  downto 0) := "00000000";  -- Amplitude control channel C
 SIGNAL R11   : std_logic_vector(7  downto 0) := "00000000";  -- Envelope period control fine tune 
 SIGNAL R12   : std_logic_vector(7  downto 0) := "00000000";  -- Envelope period control coarse tune
 SIGNAL R13   : std_logic_vector(7  downto 0) := "00000000";  -- Envelope shape/cycle control

 SIGNAL D_o      : std_logic_vector(7 downto 0);
 SIGNAL D_i      : std_logic_vector(7 downto 0);
 SIGNAL IOA_o    : std_logic_vector(7 downto 0);
 SIGNAL IOA_i    : std_logic_vector(7 downto 0);
 SIGNAL IOB_o    : std_logic_vector(7 downto 0);
 SIGNAL IOB_i    : std_logic_vector(7 downto 0);
 SIGNAL REG_ADDR : std_logic_vector(3 downto 0);

 SIGNAL RD           : std_logic;
 SIGNAL WR           : std_logic;
 SIGNAL Latch        : std_logic;
 SIGNAL BufAut       : std_logic;
 SIGNAL IOAAut       : std_logic;
 SIGNAL IOBAut       : std_logic;

 SIGNAL CLK_CORE     : std_logic;
 SIGNAL CLK_A        : std_logic;
 SIGNAL CLK_B        : std_logic; 
 SIGNAL CLK_C        : std_logic;
 SIGNAL CLK_TONE_A   : std_logic;
 SIGNAL CLK_TONE_B   : std_logic; 
 SIGNAL CLK_TONE_C   : std_logic;
 SIGNAL CLK_E        : std_logic;

 SIGNAL OUT_AMPL_E   : std_logic_vector(3 downto 0);
 SIGNAL OUT_N        : std_logic;
 SIGNAL CLK_16       : std_logic;
 SIGNAL CLK_256      : std_logic;

 SIGNAL IAnalogA     : std_logic;
 SIGNAL IAnalogB     : std_logic;
 SIGNAL IAnalogC     : std_logic;

 COMPONENT REG_ADRESSE  PORT( BUS_DATA : in     std_logic_vector(3 downto 0);
                              SEL_REG  : out    std_logic_vector(15 downto 0);
                              A8       : in     std_logic;
                              A9       : in     std_logic;
                              Latch    : in     std_logic;
                              CLK      : in     std_logic;
                              RST      : in     std_logic);
 END COMPONENT;

 COMPONENT TONE_GENERATOR  PORT ( CLK         : in     std_logic;
                                  RST         : in     std_logic; 
                                  WR          : in     std_logic;
                                  CS_COARSE   : in     std_logic;
                                  CS_FINE     : in     std_logic;
                                  DATA_COARSE : in     std_logic_vector(7 downto 0);
                                  DATA_FINE   : in     std_logic_vector(7 downto 0);
                                  OUT_TONE    : inout  std_logic );
 END COMPONENT;

 COMPONENT NOISE_GENERATOR  PORT ( CLK          : in     std_logic;
                                   RST          : in     std_logic;
                                   WR           : in     std_logic;
                                   CS           : in     std_logic;
                                   DATA         : in     std_logic_vector(4 downto 0);
                                   OUT_N        : out    std_logic);
 END COMPONENT;

 COMPONENT GEN_CLK  PORT ( CLK      : in  std_logic;
                           RST      : in  std_logic;
                           CLK_CORE : out std_logic;
                           CLK_16   : out std_logic
                          );
 END COMPONENT;

 COMPONENT MIXER    PORT ( CLK          : in     std_logic;
                           CS           : in     std_logic;
                           RST          : in     std_logic;
                           WR           : in     std_logic;
                           IN_A         : in     std_logic;
                           IN_B         : in     std_logic;
                           IN_C         : in     std_logic;
                           IN_NOISE     : in     std_logic;
                           DATA         : in     std_logic_vector(5 downto 0);
                           OUT_A        : out    std_logic;
                           OUT_B        : out    std_logic;
                           OUT_C        : out    std_logic );                           
 END COMPONENT; 

 COMPONENT GEN_ENV  PORT ( CLK_ENV      : in     std_logic;
                           DATA         : in     std_logic_vector(3 downto 0);
                           RST          : in     std_logic;
                           WR           : in     std_logic;                                    
                           CS           : in     std_logic;
                           OUT_DATA     : inout  std_logic_vector(3 downto 0));
 END COMPONENT;

 COMPONENT MANAGE_AMPLITUDE  PORT ( CLK           : in   std_logic;
                                    CLK_DAC       : in   std_logic;
                                    CLK_TONE      : in   std_logic;
                                    RST           : in   std_logic;
                                    CS            : in   std_logic;
                                    AMPLITUDE     : in   std_logic_vector(4 downto 0);
                                    AMPLITUDE_E   : in   std_logic_vector(3 downto 0);
                                    OUT_DAC       : out  std_logic );
 END COMPONENT;
 
 COMPONENT IOBUF_F_12 port (   O : out   std_logic;
                              IO : inout std_logic;
                               I : in    std_logic;
                               T : in    std_logic  );
 END COMPONENT;

 COMPONENT OBUF_F_12 port (   O : out   std_logic;
                             IO : inout std_logic;
                              I : in    std_logic;
                              T : in    std_logic  );
 END COMPONENT;

 component OBUF_F_24
 port (
   I : in std_logic;
   O : out std_logic );
 end component;

begin

-- Instantiation of sub_level modules

URA: REG_ADRESSE PORT MAP(BUS_DATA => D_i(3 downto 0), 
                          SEL_REG  => BUS_CS,                                     
                          A8       => A8, 
                          A9       => A9, 
                          Latch    => Latch, 
                          CLK      => CLOCK, 
                          RST      => RESET);

UTONE_A : TONE_GENERATOR PORT MAP( CLK         => CLOCK,
                                   RST         => RESET,
                                   WR          => WR,
                                   CS_COARSE   => BUS_CS(1),
                                   CS_FINE     => BUS_CS(0),
                                   DATA_COARSE => R1,
                                   DATA_FINE   => R0,
                                   OUT_TONE    => CLK_A);

UTONE_B : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                   RST         => RESET,
                                   WR          => WR,
                                   CS_COARSE   => BUS_CS(3),
                                   CS_FINE     => BUS_CS(2),
                                   DATA_COARSE => R3,
                                   DATA_FINE   => R2,
                                   OUT_TONE    => CLK_B);

UTONE_C : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                   RST         => RESET,
                                   WR          => WR,
                                   CS_COARSE   => BUS_CS(5),
                                   CS_FINE     => BUS_CS(4),
                                   DATA_COARSE => R5,
                                   DATA_FINE   => R4,
                                   OUT_TONE    => CLK_C);

UTONE_NOISE : NOISE_GENERATOR PORT MAP( CLK           => CLK_16,
                                        RST           => RESET,
                                        WR            => WR,
                                        CS            => BUS_CS(6),
                                        DATA          => R6(4 downto 0),
                                        OUT_N         => OUT_N);

UTONE_ENV : TONE_GENERATOR PORT MAP( CLK           => CLK_16,
                                     RST           => RESET,
                                     WR            => WR,
                                     CS_COARSE     => BUS_CS(12), 
                                     CS_FINE       => BUS_CS(11),
                                     DATA_COARSE   => R12,
                                     DATA_FINE     => R11,
                                     OUT_TONE      => CLK_E);

UCLK : GEN_CLK PORT MAP( CLK      => CLOCK, 
                         RST      => RESET,
                         CLK_CORE => CLK_CORE,
                         CLK_16   => CLK_16
                         );

UMIXER : MIXER PORT MAP ( CLK          => CLOCK,
                          CS           => BUS_CS(7),
                          RST          => RESET,
                          WR           => WR,
                          IN_A         => CLK_A,
                          IN_B         => CLK_B,
                          IN_C         => CLK_C,
                          IN_NOISE     => OUT_N,
                          DATA         => R7(5 downto 0),
                          OUT_A        => CLK_TONE_A,
                          OUT_B        => CLK_TONE_B,
                          OUT_C        => CLK_TONE_C);

UGenEnv : GEN_ENV PORT MAP( CLK_ENV      => CLK_E,
                            CS           => BUS_CS(13),                             
                            DATA         => R13(3 downto 0),
                            RST          => RESET,
                            WR           => WR,
                            OUT_DATA     => OUT_AMPL_E);

UManAmpA : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_A,
                                       RST           => RESET,
                                       CS            => BUS_CS(8),
                                       AMPLITUDE     => R8(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogA );                                       

UManAmpB : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_B,
                                       RST           => RESET,
                                       CS            => BUS_CS(9),
                                       AMPLITUDE     => R9(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogB );                                          

UManAmpC : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_C,
                                       RST           => RESET,
                                       CS            => BUS_CS(10),
                                       AMPLITUDE     => R10(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogC );                                          

PROCESS(CLOCK, RESET, CS)
 BEGIN
  if (RESET = '1') then
     Latch    <= '0';
     RD       <= '0';
     WR       <= '0';
     BufAut   <= '1';        -- High Impedance
     D_o      <= "11111111";
     IOA_o    <= "11111111";
     IOB_o    <= "11111111";
     IOAAut   <= '1';        -- High Impedance
     IOBAut   <= '1';        -- High Impedance
     REG_ADDR <= "0000";
     R0       <= "00000000"; -- for synthesis
     R1       <= "00000000"; -- for synthesis
     R2       <= "00000000"; -- for synthesis
     R3       <= "00000000"; -- for synthesis
     R4       <= "00000000"; -- for synthesis
     R5       <= "00000000"; -- for synthesis
     R6       <= "00000000"; -- for synthesis
     R7       <= "00000000"; -- for synthesis
     R8       <= "00000000"; -- for synthesis
     R9       <= "00000000"; -- for synthesis
     R10      <= "00000000"; -- for synthesis
     R12      <= "00000000"; -- for synthesis
     R13      <= "00000000"; -- for synthesis
  elsif (CLOCK'event and CLOCK = '1') then -- edge clock
    if (CS = '1') then
       D_o   <= D_i; -- ?? using by the simulation !!
       IOA_o <= IOA_i;
       IOB_o <= IOB_i;
       CASE BUSCTR IS
         WHEN "000" => BufAut      <= '1'; -- BUS_DATA is an high impedance
                       Latch       <= '0';
                       RD          <= '0';
                       WR          <= '0';
 
         WHEN "001" => BufAut      <= '0'; -- BUS_DATA is I/O, latch address
                       Latch       <= '1';  
                       RD          <= '0';
                       WR          <= '0'; 
                       REG_ADDR    <= D_i(3 downto 0);

         WHEN "010" => BufAut      <= '1'; -- BUS_DATA is an high impedance
                       Latch       <= '0';
                       RD          <= '0';
                       WR          <= '0'; 

         WHEN "011" => BufAut      <= '0'; -- BUS_DATA is I/O, RD ok, WR nok
                       Latch       <= '0';
                       RD          <= '1';
                       WR          <= '0'; 

         WHEN "100" => BufAut      <= '0'; -- BUS_DATA is I/O, latch address 
                       Latch       <= '1';
                       RD          <= '0';
                       WR          <= '0';
                       REG_ADDR    <= D_i(3 downto 0);

         WHEN "101" => BufAut      <= '1'; -- BUS_DATA is an high impedance
                       Latch       <= '0';
                       RD          <= '0';
                       WR          <= '0'; 

         WHEN "110" => BufAut      <= '0'; -- BUS_DATA is I/O, RD nok, WR ok
                       Latch       <= '0';
                       RD          <= '0';
                       WR          <= '1'; 

         WHEN "111" => BufAut      <= '0'; -- BUS_DATA is I/O, latch address
                       Latch       <= '1'; 
                       RD          <= '0';
                       WR          <= '0';   
                       REG_ADDR    <= D_i(3 downto 0);

         WHEN OTHERS => NULL;
       END CASE;

       if (RD = '1') then
          CASE REG_ADDR is
           WHEN "0000" => D_o    <= R0;
           WHEN "0001" => D_o    <= R1;
           WHEN "0010" => D_o    <= R2;
           WHEN "0011" => D_o    <= R3;
           WHEN "0100" => D_o    <= R4;
           WHEN "0101" => D_o    <= R5;
           WHEN "0110" => D_o    <= R6;
           WHEN "0111" => D_o    <= R7;
           WHEN "1000" => D_o    <= R8;
           WHEN "1001" => D_o    <= R9;
           WHEN "1010" => D_o    <= R10;
           WHEN "1011" => D_o    <= R11;
           WHEN "1100" => D_o    <= R12;
           WHEN "1101" => D_o    <= R13;
           WHEN "1110" => IOAAut <= '0'; D_o <= IOA_i;
           WHEN "1111" => IOBAut <= '0'; D_o <= IOB_i;
           WHEN OTHERS => NULL;
          END CASE;
       elsif (WR = '1') then
          CASE REG_ADDR is
           WHEN "0000" => R0     <= D_i;
           WHEN "0001" => R1     <= D_i;
           WHEN "0010" => R2     <= D_i;
           WHEN "0011" => R3     <= D_i;
           WHEN "0100" => R4     <= D_i;
           WHEN "0101" => R5     <= D_i;
           WHEN "0110" => R6     <= D_i;
           WHEN "0111" => R7     <= D_i;
           WHEN "1000" => R8     <= D_i;
           WHEN "1001" => R9     <= D_i;
           WHEN "1010" => R10    <= D_i;
           WHEN "1011" => R11    <= D_i;
           WHEN "1100" => R12    <= D_i;
           WHEN "1101" => R13    <= D_i;
           WHEN "1110" => IOAAut <= '0'; IOA_o <= D_i;
           WHEN "1111" => IOBAut <= '0'; IOB_o <= D_i;  
           WHEN OTHERS => NULL;        
          END CASE;
        end if;
    else  -- CS = 0
        BufAut <= '1';
        Latch  <= '0';
        RD     <= '0';
        WR     <= '0';
    end if;
  end if;
end process;
  
PAD_d0  : IOBUF_F_12 port map (I => D_o(0), IO => D(0), O => D_i(0), T => BufAut);
PAD_d1  : IOBUF_F_12 port map (I => D_o(1), IO => D(1), O => D_i(1), T => BufAut);
PAD_d2  : IOBUF_F_12 port map (I => D_o(2), IO => D(2), O => D_i(2), T => BufAut);
PAD_d3  : IOBUF_F_12 port map (I => D_o(3), IO => D(3), O => D_i(3), T => BufAut);
PAD_d4  : IOBUF_F_12 port map (I => D_o(4), IO => D(4), O => D_i(4), T => BufAut);
PAD_d5  : IOBUF_F_12 port map (I => D_o(5), IO => D(5), O => D_i(5), T => BufAut);
PAD_d6  : IOBUF_F_12 port map (I => D_o(6), IO => D(6), O => D_i(6), T => BufAut);
PAD_d7  : IOBUF_F_12 port map (I => D_o(7), IO => D(7), O => D_i(7), T => BufAut);

PAD_IOA0  : IOBUF_F_12 port map (I => IOA_o(0), IO => IOA(0), O => IOA_i(0), T => IOAAut);
PAD_IOA1  : IOBUF_F_12 port map (I => IOA_o(1), IO => IOA(1), O => IOA_i(1), T => IOAAut);
PAD_IOA2  : IOBUF_F_12 port map (I => IOA_o(2), IO => IOA(2), O => IOA_i(2), T => IOAAut);
PAD_IOA3  : IOBUF_F_12 port map (I => IOA_o(3), IO => IOA(3), O => IOA_i(3), T => IOAAut);
PAD_IOA4  : IOBUF_F_12 port map (I => IOA_o(4), IO => IOA(4), O => IOA_i(4), T => IOAAut);
PAD_IOA5  : IOBUF_F_12 port map (I => IOA_o(5), IO => IOA(5), O => IOA_i(5), T => IOAAut);
PAD_IOA6  : IOBUF_F_12 port map (I => IOA_o(6), IO => IOA(6), O => IOA_i(6), T => IOAAut);
PAD_IOA7  : IOBUF_F_12 port map (I => IOA_o(7), IO => IOA(7), O => IOA_i(7), T => IOAAut);


PAD_IOB0  : IOBUF_F_12 port map (I => IOB_o(0), IO => IOB(0), O => IOB_i(0), T => IOBAut);
PAD_IOB1  : IOBUF_F_12 port map (I => IOB_o(1), IO => IOB(1), O => IOB_i(1), T => IOBAut);
PAD_IOB2  : IOBUF_F_12 port map (I => IOB_o(2), IO => IOB(2), O => IOB_i(2), T => IOBAut);
PAD_IOB3  : IOBUF_F_12 port map (I => IOB_o(3), IO => IOB(3), O => IOB_i(3), T => IOBAut);
PAD_IOB4  : IOBUF_F_12 port map (I => IOB_o(4), IO => IOB(4), O => IOB_i(4), T => IOBAut);
PAD_IOB5  : IOBUF_F_12 port map (I => IOB_o(5), IO => IOB(5), O => IOB_i(5), T => IOBAut);
PAD_IOB6  : IOBUF_F_12 port map (I => IOB_o(6), IO => IOB(6), O => IOB_i(6), T => IOBAut);
PAD_IOB7  : IOBUF_F_12 port map (I => IOB_o(7), IO => IOB(7), O => IOB_i(7), T => IOBAut);

PAD_ANALOGA : OBUF_F_24 port map( I => IAnalogA, O => AnalogA);
PAD_ANALOGB : OBUF_F_24 port map( I => IAnalogB, O => AnalogB);
PAD_ANALOGC : OBUF_F_24 port map( I => IAnalogC, O => AnalogC);
end Behavioral;
